module texpzl.texpzl;

import derelict.sdl.sdl;
import derelict.sdl.image;
import std.stdio, std.path, std.exception;
import std.string;
import std.zlib;
import std.traits;
import std.algorithm;


uint nearPow2(uint x) {
    uint i = 1;
    while (i < x) i *= 2;
    return i;
}


void Texpzl_init() {
    DerelictSDL.load();
    DerelictSDLImage.load();
    enforce(SDL_Init(SDL_INIT_VIDEO) == 0);
    enforce(IMG_Init(IMG_INIT_PNG) & IMG_INIT_PNG);
}

void Texpzl_quit() {
    IMG_Quit();
    SDL_Quit();
}

struct Assign {
    // assign
    bool valid;
    int x, y;

    // option
    int padding;
    
    int width() @property const { return (roll ? orig_height : orig_width) + padding*2; }
    int height() @property const { return (roll ? orig_width : orig_height) + padding*2; }

    // image
    string name;
    int orig_width, orig_height;
    bool roll;
    SDL_Surface* surface;

    // for sort
    int opCmp(ref const Assign b) const {
        return b.height - this.height;
    }
}

struct Images {
    Assign[] assigns;
    
    void addImage(string file, string name, int padding=0) {
        Assign image;
        image.name = name;
        image.surface = IMG_Load(file.toStringz);
        image.orig_width  = image.surface.w;
        image.orig_height = image.surface.h;
        image.padding = padding;
        assigns ~= image;
    }
}

struct Rect {
    int x, y;
    int width, height;
}

struct PixelsRGBA {
    int width, height;
    enum bytes_per_pixel = 4;
    ubyte[] data;
    ubyte[] compressed_data;

    void resize(int w, int h) {
        width  = w;
        height = h;
        foreach (ref v; data[0 .. min(w*h, data.length)])
            v = 0;
        data.length = w * h * bytes_per_pixel + (h * 1);
    }
    
    void set(int x, int y, ubyte r, ubyte g, ubyte b, ubyte a) {
        if (x < 0 || y < 0 || x >= width || y >= height)
            return;
        // + (y + 1)はスキャンラインごとのフィルタ設定バイトを飛ばすため
        ubyte* p = &data[(width * y + x) * bytes_per_pixel + (y + 1)];
        p[0] = r;
        p[1] = g;
        p[2] = b;
        p[3] = a;
    }

    void compress() {
        compressed_data = cast(ubyte[])data.compress();
    }
}

// PixelsRGBAにコピーで配置する。
void doAssign(ref PixelsRGBA dest, Assign assign) {
    if (SDL_MUSTLOCK(assign.surface)) SDL_LockSurface(assign.surface);
    scope(exit) if (SDL_MUSTLOCK(assign.surface)) SDL_UnlockSurface(assign.surface);
    
    auto basex = assign.x+assign.padding, basey = assign.y+assign.padding;
    
    foreach (y; 0 .. assign.orig_height) {
        foreach (x; 0 .. assign.orig_width) {
            auto pixel = getpixel(assign.surface, x, y);
            ubyte r, g, b, a;
            SDL_GetRGBA(pixel, assign.surface.format, &r, &g, &b, &a);
            if (assign.roll)
                dest.set(basex+y, basey+(assign.height-x-1-(assign.padding*2)), r, g, b, a);
            else
                dest.set(basex+x, basey+y, r, g, b, a);
        }
    }
}

// from SDL manual
uint getpixel(SDL_Surface *surface, int x, int y) {
    auto bpp = surface.format.BytesPerPixel;
    auto p = cast(ubyte*)surface.pixels + y * surface.pitch + x * bpp;

    switch (bpp) {
    case 1:
        return *p;
    case 2:
        return *cast(ushort*)p;
    case 3:
        if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
            return p[0] << 16 | p[1] << 8 | p[2];
        else
            return p[0] | p[1] << 8 | p[2] << 16;
    case 4:
        return *cast(uint*)p;
    default:
        return 0;
    }
}

T reverseByte(T)(T t) {
    auto a = (cast(ubyte*)&t)[0 .. T.sizeof].dup.reverse;
    return *cast(T*)a.ptr;
}
T identity(T)(T t) {
    return t;
}

import std.system;
static if (std.system.endian == Endian.bigEndian) {
    alias indentity toBigEndian;
}
else static if (std.system.endian == Endian.littleEndian) {
    alias reverseByte toBigEndian;
}


struct PNG {
    enum ubyte[8] ID = [0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A];
    enum ChunkSize = 12;
    align(1) struct Chunk(T) {
        uint size;
        ubyte[4] name = cast(ubyte[4])T.stringof;
        T body_;
        alias body_ this;
        uint crc32;
    }
    align(1) struct IHDR {
        uint width, height;
        ubyte depth;
        ubyte color;
        ubyte compress;
        ubyte filter;
        ubyte interlace;
    }
    align(1) struct IDAT {
        ubyte[] data;
    }
    align(1) struct IEND {
        ubyte[0] data;
    }

    static void save(ref PixelsRGBA src, string file) {
        Chunk!IHDR ihdr;
        ihdr.size = toBigEndian(IHDR.sizeof);
        ihdr.width  = toBigEndian(src.width);
        ihdr.height = toBigEndian(src.height);
        ihdr.depth = 8;
        ihdr.color = 6;
        Chunk!IDAT idat;
        idat.data = src.compressed_data;
        idat.size = toBigEndian(idat.data.length);
        Chunk!IEND iend;
        iend.size = 0;

        ihdr.crc32 = crc32(crc32(0, ihdr.name), (cast(ubyte*)&ihdr.body_)[0 .. ihdr.body_.sizeof]);
        ihdr.crc32 = toBigEndian(ihdr.crc32);
        idat.crc32 = crc32(crc32(0, idat.name), idat.data);
        idat.crc32 = toBigEndian(idat.crc32);
        iend.crc32 = crc32(0, iend.name);
        iend.crc32 = toBigEndian(iend.crc32);
        
        auto f = File(file, "w");
        void write(T)(T t) {
            static if (is(T == struct)) {
                foreach (ref v; t.tupleof) {
                    write(v);
                }
            }
            else static if (isArray!T) {
                foreach (ref v; t) {
                    write(v);
                }
            }
            else {
                f.rawWrite((cast(ubyte*)&t)[0 .. T.sizeof]);
            }
        }
        write(ID);
        write(ihdr);
        
        write(idat.size);
        write(idat.name);
        f.rawWrite(idat.data);
        write(idat.crc32);
        write(iend);
    }
}


void saveAssigns(string file, int width, int height, Assign[] assigns, bool binmode) {
    if (binmode) {
        auto assignf = File(file, "wb");
        assignf.rawWrite(cast(ubyte[])("TEXP".dup));
        writeln(*cast(int*)("TEXP".ptr));
        assignf.rawWrite([width, width, assigns.length]);
        foreach (a; assigns) {
            assignf.rawWrite([a.name.length]);
            assignf.rawWrite(cast(ubyte[])(a.name.dup));
            assignf.rawWrite([a.x+a.padding, a.y+a.padding, a.orig_width, a.orig_height, a.roll ? 1 : 0]);
        }
    }
    else {
        auto assignf = File(file, "w");
        assignf.writeln(width, ",", width, ",", assigns.length);
        foreach (a; assigns) {
            assignf.writeln(a.name, ",", a.x+a.padding, ",", a.y+a.padding, ",",
                            a.orig_width, ",", a.orig_height, ",", a.roll?1:0);
        }
    }
}



